from ..utils import Scraper
from bs4 import BeautifulSoup
from calendar import monthrange
from collections import OrderedDict
import datetime
import re
import time


class Shuttle:
    """A scraper for UofT's shuttle bus schedule.

    The schedule is located at https://m.utm.utoronto.ca/shuttle.php.
    """

    host = 'https://m.utm.utoronto.ca/shuttleByDate.php?year=%s&month=%s&day=%s'

    building_ids = {
        'Instructional Centre Layby': '334',
        'Hart House': '002',
        'Deerfield Hall North Layby': '340'
    }

    @staticmethod
    def scrape(location='.', month=None):
        """Update the local JSON files for this scraper."""

        Scraper.logger.info('Shuttle initialized.')

        now = datetime.datetime.now()
        year = now.strftime('%Y')
        month = now.strftime('%m') if month is None else str(month).zfill(2)
        days = monthrange(int(year), int(month))[1]

        Scraper.logger.info('Fetching schedules for {0}-{1}-01 to {0}-{1}-{2}.'.format(year, month, days))

        for day in range(1, days + 1):
            html = Scraper.get_html(Shuttle.host % (year, month, day))
            schedule = Shuttle.parse_schedule_html(html)

            Scraper.save_json(schedule, location, schedule['date'])

        Scraper.logger.info('Shuttle completed.')

    @staticmethod
    def parse_schedule_html(html):
        """Create JSON object from the HTML page."""

        soup = BeautifulSoup(html, 'html.parser')

        # Get date
        date = time.strftime('%Y-%m-%d', time.strptime(soup.find('h2').get_text().strip(), '%b %d %Y'))

        # Get route data
        routes = {}
        _routes = soup.find(id='chooseRoute')
        if _routes:
            for _route in _routes.find_all('option'):
                _route_id = _route.get('value')
                _route_times = soup.find(id=_route_id).find_all('li')

                route_name, route_location = _route.get_text().strip().split(' @ ')
                route_id = re.sub('\.|ROUTE| ', '', route_name.upper())

                times = []
                for _route_time in _route_times:
                    _route_time_text = _route_time.get_text().strip().lower()
                    _route_time_clean = re.sub('\*.*\*', '', _route_time_text).strip()

                    time_rush_hour = 'rush hour' in _route_time_text
                    time_no_overload = 'no overload' in _route_time_text

                    times.append(OrderedDict([
                        ('time', '%sT%s:00-04:00' % (
                            date,
                            time.strftime('%H:%M %p', time.strptime(_route_time_clean, '%I:%M %p'))[:-3]
                        )),
                        ('rush_hour', time_rush_hour),
                        ('no_overload', time_no_overload)
                    ]))

                # TODO: fetch this dynamically
                route_building_id = Shuttle.building_ids[route_location] if route_location in Shuttle.building_ids else ''

                route_stops = OrderedDict([
                    ('location', route_location),
                    ('building_id', route_building_id),
                    ('times', times)
                ])

                if route_id in routes.keys():
                    routes[route_id]['stops'].append(route_stops)
                else:
                    routes[route_id] = OrderedDict([
                        ('id', route_id),
                        ('name', route_name),
                        ('stops', [route_stops])
                    ])

        return OrderedDict([
            ('date', date),
            ('routes', [v for k, v in routes.items()])
        ])



import logging
import os
import sys

from .scrapers.courses import Courses

from .scrapers.buildings import Buildings

from .scrapers.textbooks import Textbooks

from .scrapers.food import Food

from .scrapers.calendar import Calendar
from .scrapers.calendar.utsg import UTSGCalendar
from .scrapers.calendar.utm import UTMCalendar
from .scrapers.calendar.utsc import UTSCCalendar

from .scrapers.timetable import Timetable
from .scrapers.timetable.utm import UTMTimetable
from .scrapers.timetable.utsc import UTSCTimetable
from .scrapers.timetable.utsg import UTSGTimetable

from .scrapers.exams import Exams
from .scrapers.exams.utsg import UTSGExams
from .scrapers.exams.utm import UTMExams
from .scrapers.exams.utsc import UTSCExams

from .scrapers.athletics import Athletics
from .scrapers.athletics.utsg import UTSGAthletics
from .scrapers.athletics.utm import UTMAthletics
from .scrapers.athletics.utsc import UTSCAthletics

from .scrapers.parking import Parking

from .scrapers.shuttle import Shuttle


class NullHandler(logging.Handler):
    def emit(self, record):
        pass

logging.getLogger('uoftscrapers').addHandler(NullHandler())



from ..utils import Scraper
from bs4 import BeautifulSoup
from collections import OrderedDict
from queue import Queue
from threading import Thread, Lock
from time import time, sleep
import http.cookiejar
import json
import logging
import os
import re
import requests
import sys


class Courses:
    """A scraper for UofT's Course Finder web service.

    Course Finder is located at http://coursefinder.utoronto.ca/.
    """

    host = 'http://coursefinder.utoronto.ca/course-search/search'
    cookies = http.cookiejar.CookieJar()
    s = requests.Session()
    threads = 32

    @staticmethod
    def scrape(location='.'):
        """Update the local JSON files for this scraper."""

        Scraper.logger.info('Courses initialized.')

        urls = Courses.search()
        total = len(urls)

        ts = time()
        queue = Queue()

        for x in range(Courses.threads):
            worker = CourseFinderWorker(queue)
            worker.daemon = True
            worker.start()

        Scraper.logger.info('Queued %d courses.' % total)
        for x in urls:
            course_id = re.search('offImg(.*)', x[0]).group(1).split('"')[0]
            url = '%s/courseSearch/coursedetails/%s' % (
                Courses.host,
                course_id
            )
            queue.put((course_id, url, total))

        queue.join()

        Scraper.logger.info('Took %.2fs to retreive course info.' % (
            time() - ts
        ))

        for course in CourseFinderWorker.all_courses:
            if course != False:
                Scraper.save_json(course, location, course['id'])

        Scraper.logger.info('Courses completed.')

    @staticmethod
    def search(query='', requirements=''):
        """Perform a search and return the data as a dict."""

        url = '%s/courseSearch/course/search' % Courses.host

        data = {
            'queryText': query,
            'requirements': requirements,
            'campusParam': 'St. George,Scarborough,Mississauga'
        }

        # Keep trying to get data until a proper response is given
        json = None
        while json is None:
            try:
                r = Courses.s.get(url, params=data,
                    cookies=Courses.cookies)
                if r.status_code == 200:
                    json = r.json()
                else:
                    sleep(0.5)
            except requests.exceptions.Timeout:
                continue

        return json['aaData']

    @staticmethod
    def get_course_html(url):
        """Update the locally stored course pages."""

        html = None
        while html is None:
            try:
                r = Courses.s.get(url, cookies=Courses.cookies)
                if r.status_code == 200:
                    html = r.text
            except (requests.exceptions.Timeout,
                    requests.exceptions.ConnectionError):
                continue

        return html.encode('utf-8')

    @staticmethod
    def parse_course_html(course_id, html):
        """Create JSON files from the HTML pages downloaded."""

        if "The course you are trying to access does not exist" in \
                html.decode('utf-8'):
            return False

        soup = BeautifulSoup(html, "html.parser")

        # Things that appear on all courses

        title = soup.find(id="u19")
        title_name = title.find_all("span",
                                    class_="uif-headerText-span")[0].get_text()

        course_code = course_id[:-5]

        course_name = title_name[10:]

        division = soup.find(id="u23").find_all("span", id="u23")[0] \
            .get_text().strip()

        description = soup.find(id="u32").find_all("span", id="u32")[0] \
            .get_text().strip()

        department = soup.find(id="u41").find_all("span", id="u41")[0] \
            .get_text().strip()

        course_level = soup.find(id="u86").find_all("span", id="u86")[0] \
            .get_text().strip()
        course_level = course_level[:3]
        course_level = int(course_level)

        campus = soup.find(id="u149").find_all("span", id="u149")[0] \
            .get_text().strip()

        if campus == "St. George":
            campus = "UTSG"
        elif campus == "Mississauga":
            campus = "UTM"
        elif campus == "Scarborough":
            campus = "UTSC"

        term = soup.find(id="u158").find_all("span", id="u158")[0] \
            .get_text().strip()

        # Things that don't appear on all courses

        as_breadth = soup.find(id="u122")
        breadths = []
        if as_breadth is not None:
            as_breadth = as_breadth.find_all("span", id="u122")[0] \
                .get_text().strip()
            for ch in as_breadth:
                if ch in "12345":
                    breadths.append(int(ch))

        breadths = sorted(breadths)

        exclusions = soup.find(id="u68")
        if exclusions is not None:
            exclusions = exclusions.find_all("span", id="u68")[0] \
                .get_text().strip()
        else:
            exclusions = ""

        prereq = soup.find(id="u50")
        if prereq is not None:
            prereq = prereq.find_all("span", id="u50")[0].get_text().strip()
        else:
            prereq = ""

        # Meeting Sections

        meeting_table = soup.find(id="u172")

        trs = []
        if meeting_table is not None:
            trs = meeting_table.find_all("tr")

        sections = []

        for tr in trs:
            tds = tr.find_all("td")
            if len(tds) > 0:
                code = tds[0].get_text().strip()

                raw_times = tds[1].get_text().replace(
                    'Alternate week', '').strip().split(" ")
                times = []
                for i in range(0, len(raw_times) - 1, 2):
                    times.append(raw_times[i] + " " + raw_times[i + 1])

                instructors = BeautifulSoup(str(tds[2]).replace("<br>", "\n"),
                    "html.parser")
                instructors = instructors.get_text().split("\n")
                instructors = \
                    list(filter(None, [x.strip() for x in instructors]))

                raw_locations = tds[3].get_text().strip().split(" ")
                locations = []
                for i in range(0, len(raw_locations) - 1, 2):
                    locations.append(
                        raw_locations[i] + " " + raw_locations[i + 1])

                class_size = tds[4].get_text().strip()

                time_data = []
                for i in range(len(times)):
                    info = times[i].split(" ")
                    day = info[0]
                    hours = info[1].split("-")

                    location = ""
                    try:
                        location = locations[i]
                    except IndexError:
                        location = ""

                    for i in range(len(hours)):
                        x = hours[i].split(':')
                        hours[i] = int(x[0]) + (int(x[1])/60)

                    time_data.append(OrderedDict([
                        ("day", day),
                        ("start", hours[0]),
                        ("end", hours[1]),
                        ("duration", hours[1] - hours[0]),
                        ("location", location)
                    ]))

                code = code.split(" ")
                code = code[0][0] + code[1]

                data = OrderedDict([
                    ("code", code),
                    ("instructors", instructors),
                    ("times", time_data),
                    ("size", int(class_size)),
                    ("enrolment", 0)
                ])

                sections.append(data)

        # Dictionary creation
        course = OrderedDict([
            ("id", course_id),
            ("code", course_code),
            ("name", course_name),
            ("description", description),
            ("division", division),
            ("department", department),
            ("prerequisites", prereq),
            ("exclusions", exclusions),
            ("level", course_level),
            ("campus", campus),
            ("term", term),
            ("breadths", breadths),
            ("meeting_sections", sections)
        ])

        return course


class CourseFinderWorker(Thread):

    all_courses = []
    done = 0
    lock = Lock()

    def __init__(self, queue):
        Thread.__init__(self)
        self.queue = queue

    def run(self):
        while True:
            course_id, url, total = self.queue.get()
            html = Courses.get_course_html(url)
            course = Courses.parse_course_html(course_id, html)

            CourseFinderWorker.lock.acquire()
            CourseFinderWorker.all_courses.append(course)
            CourseFinderWorker.done += 1
            Scraper.flush_percentage(CourseFinderWorker.done / total)
            CourseFinderWorker.lock.release()

            self.queue.task_done()



from ..utils import Scraper, LayersScraper
from bs4 import BeautifulSoup
from collections import OrderedDict
from pprint import pprint
import json
import requests


class Parking:
    """A scraper for parking lots / bicycle racks.

    Parking data is located at http://map.utoronto.ca
    """

    host = 'http://map.utoronto.ca/'
    indices = {
        'utsg': [4, 1],
        'utm': 6,
        'utsc': 5
    }
    s = requests.Session()

    @staticmethod
    def scrape(location='.'):
        """Update the local JSON files for this scraper."""

        Scraper.logger.info('Parking initialized.')

        data = LayersScraper.get_layers_json('utsg')

        # UTSG car parking
        for entry in data[Parking.indices['utsg'][0]]['markers']:
            _id = str(entry['id']).zfill(4)
            title = LayersScraper.get_value(entry, 'title')
            building_id = LayersScraper.get_value(entry, 'building_code')
            _type = 'car'
            access = LayersScraper.get_value(entry, 'access')
            description = BeautifulSoup(
                LayersScraper.get_value(entry, 'desc').strip(),
                'html.parser').text
            lat = LayersScraper.get_value(entry, 'lat', True)
            lng = LayersScraper.get_value(entry, 'lng', True)
            address = LayersScraper.get_value(entry, 'address')

            doc = OrderedDict([
                ('id', _id),
                ('title', title),
                ('building_id', building_id),
                ('campus', 'UTSG'),
                ('type', _type),
                ('description', description),
                ('lat', lat),
                ('lng', lng),
                ('address', address)
            ])

            Scraper.save_json(doc, location, _id)

        # UTSG bicycle parking
        for entry in data[Parking.indices['utsg'][1]]['markers']:
            if 64 in entry['attribs']:
                # Ignore Bikeshare (third party)
                continue

            _id = str(entry['id']).zfill(4)
            title = LayersScraper.get_value(entry, 'title')
            building_id = LayersScraper.get_value(entry, 'building_code')
            _type = 'bicycle'
            access = LayersScraper.get_value(entry, 'access')
            description = BeautifulSoup(
                LayersScraper.get_value(entry, 'desc').strip(),
                'html.parser').text
            if len(description) > 0 and description[-1] == '.':
                description = description[:-1]

            lat = LayersScraper.get_value(entry, 'lat', True)
            lng = LayersScraper.get_value(entry, 'lng', True)
            address = LayersScraper.get_value(entry, 'address')

            doc = OrderedDict([
                ('id', _id),
                ('title', title),
                ('building_id', building_id),
                ('campus', 'UTSG'),
                ('type', _type),
                ('description', description),
                ('lat', lat),
                ('lng', lng),
                ('address', address)
            ])

            Scraper.save_json(doc, location, _id)

        # UTM / UTSC car parking
        for campus in ('utm', 'utsc'):
            data = LayersScraper.get_layers_json(campus)

            for entry in data[Parking.indices[campus]]['markers']:
                if 'parking' not in LayersScraper.get_value(entry, 'slug'):
                    continue

                _id = str(entry['id']).zfill(4)
                title = LayersScraper.get_value(entry, 'title')
                building_id = LayersScraper.get_value(entry, 'building_code')
                _type = 'car'
                description = BeautifulSoup(
                    LayersScraper.get_value(entry, 'desc').strip(),
                    'html.parser').text
                lat = LayersScraper.get_value(entry, 'lat', True)
                lng = LayersScraper.get_value(entry, 'lng', True)
                address = LayersScraper.get_value(entry, 'address')

                doc = OrderedDict([
                    ('id', _id),
                    ('title', title),
                    ('building_id', building_id),
                    ('campus', campus.upper()),
                    ('type', _type),
                    ('description', description),
                    ('lat', lat),
                    ('lng', lng),
                    ('address', address)
                ])

                with open('%s/%s.json' % (location, _id), 'w') as fp:
                    json.dump(doc, fp)

        Scraper.logger.info('Parking completed.')



from ..utils import Scraper
from bs4 import BeautifulSoup
from collections import OrderedDict
from operator import itemgetter
from pprint import pprint
from queue import Queue
from threading import Thread, Lock
from time import time
import logging
import os
import re
import requests
import sys


class Textbooks:
    """A scraper for UofT's book store.

    UofT Book Store is located at http://uoftbookstore.com/.
    """

    host = 'http://uoftbookstore.com'
    threads = 32

    @staticmethod
    def scrape(location='.'):
        """Update the local JSON files for this scraper."""

        Scraper.logger.info('Textbooks initialized.')

        terms = Textbooks.retrieve_terms()
        departments = Textbooks.retrieve_departments(terms)

        # Get course info
        ts = time()
        queue = Queue()

        for x in range(Textbooks.threads):
            worker = CoursesWorker(queue)
            worker.daemon = True
            worker.start()

        total = len(departments)
        Scraper.logger.info('Queued %d departments. (1/3)' % total)
        for department in departments:
            queue.put((department, total))

        queue.join()
        Scraper.logger.info('Took %.2fs to retreive course info.' % (
            time() - ts
        ))

        # Get section info
        ts = time()
        queue = Queue()

        for x in range(Textbooks.threads):
            worker = SectionsWorker(queue)
            worker.daemon = True
            worker.start()

        total = len(CoursesWorker.all_courses)
        Scraper.logger.info('Queued %d courses. (2/3)' % total)
        for course in CoursesWorker.all_courses:
            queue.put((course, total))

        queue.join()
        Scraper.logger.info('Took %.2fs to retreive section info.' % (
            time() - ts
        ))

        # Get book info
        ts = time()
        queue = Queue()

        for x in range(Textbooks.threads):
            worker = BooksWorker(queue)
            worker.daemon = True
            worker.start()

        total = len(SectionsWorker.all_sections)
        Scraper.logger.info('Queued %d sections. (3/3)' % total)
        for section in SectionsWorker.all_sections:
            queue.put((section, total))

        queue.join()
        Scraper.logger.info('Took %.2fs to retreive book info.' % (
            time() - ts
        ))

        books = list(BooksWorker.all_books.values())

        # Sort books and dump the files
        for book in books:
            book['courses'] = sorted(book['courses'], key=itemgetter('id'))
            for i in range(len(book['courses'])):
                book['courses'][i]['meeting_sections'] = \
                sorted(book['courses'][i]['meeting_sections'],
                    key=itemgetter('code'))

            Scraper.save_json(book, location, book['id'])

        Scraper.logger.info('Textbooks completed.')

    @staticmethod
    def retrieve_terms():
        r = requests.get('%s/buy_courselisting.asp' % Textbooks.host)

        listing = BeautifulSoup(r.text, "html.parser")
        terms = listing.find(id='fTerm').find_all('option')[1:]

        accepted_terms = []
        for term in terms:
            val = term.get_text()
            if val.startswith('ST GEORGE') or val.startswith('MISSISSAUGA') \
                or val.startswith('SCARBOROUGH'):
                accepted_terms.append(term)

        return accepted_terms

    @staticmethod
    def retrieve_departments(terms):
        all_departments = []

        for term in terms:
            term_name = term.get_text()
            m = re.search('(\d{5})', term_name)
            session = m.group(0)

            campus, term_id = term.get('value').split('|')
            payload = {
                'control': 'campus',
                'campus': campus,
                'term': term_id,
                't': int(round(time() * 1000))
            }
            headers = {
                'Referer': '%s/buy_courselisting.asp' % Textbooks.host
            }

            r = requests.get('%s/textbooks_xml.asp' % Textbooks.host,
                params=payload, headers=headers)

            departments = BeautifulSoup(r.text, "xml").find_all('department')
            for department in departments:
                all_departments.append({
                    'dept_id': department.get('id'),
                    'dept_name': department.get('name').title(),
                    'term_id': term_id,
                    'session': session
                })

            Scraper.logger.info('Retreived department info from %s.' % term_name)

        return all_departments

    @staticmethod
    def retrieve_courses(department):
        all_courses = []

        payload = {
            'control': 'department',
            'dept': department['dept_id'],
            'term': department['term_id'],
            't': int(round(time() * 1000))
        }
        headers = {
            'Referer': '%s/buy_courselisting.asp' % Textbooks.host
        }

        r = requests.get('%s/textbooks_xml.asp' % Textbooks.host,
            params=payload, headers=headers)

        courses = BeautifulSoup(r.text, "xml").find_all('course')
        for course in courses:
            all_courses.append({
                'course_id': course.get('id'),
                'course_name': course.get('name'),
                'term_id': department['term_id'],
                'session': department['session']
            })

        return all_courses

    @staticmethod
    def retrieve_sections(course):
        all_sections = []

        payload = {
            'control': 'course',
            'course': course['course_id'],
            'term': course['term_id'],
            't': int(round(time() * 1000))
        }
        headers = {
            'Referer': '%s/buy_courselisting.asp' % Textbooks.host
        }

        r = requests.get('%s/textbooks_xml.asp' % Textbooks.host,
            params=payload, headers=headers)

        sections = BeautifulSoup(r.text, "xml").find_all('section')
        for section in sections:
            all_sections.append({
                'section_id': section.get('id'),
                'section_code': section.get('name'),
                'section_instructor': section.get('instructor'),
                'course_code': course['course_name'],
                'session': course['session']
            })

        return all_sections

    @staticmethod
    def retrieve_books(section):
        all_books = []

        payload = {
            'control': 'section',
            'section': section['section_id'],
            't': int(round(time() * 1000))
        }
        headers = {
            'Referer': '%s/buy_courselisting.asp' % Textbooks.host
        }

        r = requests.get('%s/textbooks_xml.asp' % Textbooks.host,
            params=payload, headers=headers)

        soup = BeautifulSoup(r.text, "html.parser")
        books = soup.find_all('tr', { 'class': 'book' })

        if books == None:
            done += 1
            return []

        for book in books:
            if len(book.get_text().strip()) == 0:
                continue

            image = book.find(class_='book-cover').img.get('src')
            image = 'http://uoftbookstore.com/%s' % image
            image = image.replace('Size=M', 'Size=L')

            # This doesn't mean "avoid textbooks with no image"
            # This is a case when the textbook is called "No required text"
            if 'not_available_' in image:
                continue

            book_id = book.find(class_='product-field-pf_id').get('value')

            url = '%s/buy_book_detail.asp?pf_id=%s' % (Textbooks.host, book_id)

            title = Scraper.get_text_from_class(book, 'book-title')

            edition = Scraper.get_text_from_class(book, 'book-edition')
            if len(edition) > 0:
                edition = ''.join(list(filter(str.isdigit, edition)))
                try:
                    edition = int(edition)
                except ValueError:
                    edition = 1
            if edition == '' or 0:
                edition = 1

            author = Scraper.get_text_from_class(book, 'book-author')
            m = re.search('([\d]+[E]?)', author)
            if m != None:
                junk = m.group(0)
                author = author.replace(junk, '').strip()

            isbn = Scraper.get_text_from_class(book, 'isbn')
            requirement = Scraper.get_text_from_class(book, 'book-req')
            requirement = requirement.lower()

            price = Scraper.get_text_from_class(book, 'book-price-list')
            try:
                price = float(price[1:])
            except ValueError:
                price = 0

            instructor = section['section_instructor'].split(',')
            if len(instructor) == 2:
                instructor = '%s %s' % (
                    instructor[0][:1],
                    instructor[1].strip()
                )
                instructor = instructor.strip()
            else:
                instructor = ''

            instructors = [instructor]
            if len(instructor) == 0:
                instructors = []

            meeting_sections = [OrderedDict([
                ("code", section['section_code']),
                ("instructors", instructors)
            ])]

            course_id = '%s%s' % (section['course_code'], section['session'])

            courses = [OrderedDict([
                ("id", course_id),
                ("code", section['course_code']),
                ("requirement", requirement),
                ("meeting_sections", meeting_sections)
            ])]

            textbook = OrderedDict([
                ("id", book_id),
                ("isbn", isbn),
                ("title", title),
                ("edition", edition),
                ("author", author),
                ("image", image),
                ("price", price),
                ("url", url),
                ("courses", courses)
            ])

            all_books.append(textbook)

        return all_books


class CoursesWorker(Thread):

    all_courses = []
    done = 0
    lock = Lock()

    def __init__(self, queue):
        Thread.__init__(self)
        self.queue = queue

    def run(self):
        while True:
            department, total = self.queue.get()
            courses = Textbooks.retrieve_courses(department)

            CoursesWorker.lock.acquire()
            CoursesWorker.all_courses += courses
            CoursesWorker.done += 1
            Scraper.flush_percentage(CoursesWorker.done / total)
            CoursesWorker.lock.release()

            self.queue.task_done()


class SectionsWorker(Thread):

    all_sections = []
    done = 0
    lock = Lock()

    def __init__(self, queue):
        Thread.__init__(self)
        self.queue = queue

    def run(self):
        while True:
            course, total = self.queue.get()
            sections = Textbooks.retrieve_sections(course)

            SectionsWorker.lock.acquire()
            SectionsWorker.all_sections += sections
            SectionsWorker.done += 1
            Scraper.flush_percentage(SectionsWorker.done / total)
            SectionsWorker.lock.release()

            self.queue.task_done()


class BooksWorker(Thread):

    all_books = {}
    done = 0
    lock = Lock()

    def __init__(self, queue):
        Thread.__init__(self)
        self.queue = queue

    def run(self):
        while True:
            section, total = self.queue.get()
            books = Textbooks.retrieve_books(section)

            BooksWorker.lock.acquire()
            for book in books:
                if book['id'] in BooksWorker.all_books:
                    index = -1
                    for i in range(len(BooksWorker.all_books[book['id']]['courses'])):
                        if BooksWorker.all_books[book['id']]['courses'][i]['id'] == book['courses'][0]['id']:
                            index = i
                            break
                    if index >= 0:
                        BooksWorker.all_books[book['id']]['courses'][index]['meeting_sections'] += book['courses'][0]['meeting_sections']
                    else:
                        BooksWorker.all_books[book['id']]['courses'] += book['courses']
                else:
                    BooksWorker.all_books[book['id']] = book
            BooksWorker.done += 1
            Scraper.flush_percentage(BooksWorker.done / total)
            BooksWorker.lock.release()

            self.queue.task_done()



from ..utils import Scraper, LayersScraper
from bs4 import BeautifulSoup
from collections import OrderedDict
from decimal import *
import json
import os
import re
import requests


class Buildings:
    """A scraper for UofT's buildings.

    UofT Map is located at http://map.utoronto.ca/.
    """

    host = 'http://map.utoronto.ca/'
    campuses = ['utsg', 'utm', 'utsc']
    s = requests.Session()

    @staticmethod
    def scrape(location='.'):
        """Update the local JSON files for this scraper."""

        Scraper.logger.info('Buildings initialized.')

        for campus in Buildings.campuses:
            data = Buildings.get_map_json(campus)
            regions = Buildings.get_regions_json(campus)['buildings']

            for building in data['buildings']:
                _id = building['id']
                code = building['code']
                name = building['title']
                short_name = LayersScraper.get_value(building, 'short_name')
                lat = LayersScraper.get_value(building, 'lat', True)
                lng = LayersScraper.get_value(building, 'lng', True)

                street = ' '.join(filter(None,
                    LayersScraper.get_value(building, 'street').split(' ')))
                city = LayersScraper.get_value(building, 'city')
                province = LayersScraper.get_value(building, 'province')
                country = LayersScraper.get_value(building, 'country')
                postal = LayersScraper.get_value(building, 'postal')

                polygon = []
                for region in regions:
                    if region['id'] == _id:
                        lat_lng = region['center_point']
                        if lat_lng:
                            lat_lng = lat_lng[1:-2].split(', ')
                            if len(lat_lng) == 2:
                                lat = float(lat_lng[0])
                                lng = float(lat_lng[1])
                            polygon = region['points']

                doc = OrderedDict([
                    ('id', _id),
                    ('code', code),
                    ('name', name),
                    ('short_name', short_name),
                    ('campus', campus.upper()),
                    ('address', OrderedDict([
                        ('street', street),
                        ('city', city),
                        ('province', province),
                        ('country', country),
                        ('postal', postal)
                    ])),
                    ('lat', lat),
                    ('lng', lng),
                    ('polygon', polygon)
                ])

                Scraper.save_json(doc, location, _id)

        Scraper.logger.info('Buildings completed.')

    @staticmethod
    def get_map_json(campus):
        """Retrieve the JSON structure from host."""

        Scraper.logger.info('Scraping %s.' % campus.upper())

        Buildings.s.get(Buildings.host)

        headers = { 'Referer': Buildings.host }
        html = Buildings.s.get('%s%s%s' % (
            Buildings.host,
            'data/map/',
            campus
        ), headers=headers).text

        data = json.loads(html)
        return data

    @staticmethod
    def get_regions_json(campus):
        """Retrieve the JSON structure from host."""

        Buildings.s.get(Buildings.host)

        headers = { 'Referer': Buildings.host }
        html = Buildings.s.get('%s%s%s' % (
            Buildings.host,
            'data/regions/',
            campus
        ), headers=headers).text

        data = json.loads(html)
        return data



from ..utils import Scraper
from bs4 import BeautifulSoup
from collections import OrderedDict
import json
import os
import requests


class UTSGCalendar:

    host = 'http://www.artsandscience.utoronto.ca/ofr/calendar/'

    @staticmethod
    def scrape(location='.'):
        Scraper.logger.info('UTSGCalendar initialized.')
        Scraper.logger.info('Not implemented.')
        Scraper.logger.info('UTSGCalendar completed.')



from ..utils import Scraper
from .utsg import UTSGAthletics
from .utm import UTMAthletics
from .utsc import UTSCAthletics


class Athletics:
    @staticmethod
    def scrape(location='.'):
        Scraper.logger.info('Athletics initialized.')
        UTSGAthletics.scrape(location)
        UTMAthletics.scrape(location)
        UTSCAthletics.scrape(location)
        Scraper.logger.info('Athletics completed.')
